//--------------------------------------------------------------------
//
//  RENLIB - Renju Library Program
//
//  Author : Frank Arkbo
//
//----------------------------------------------------------------------

const char* version = "HP-RENLIB 1997-02-23";

#include <conio.h>
#include <dir.h>
#include <dos.h>
#include <float.h>
#include <graphics.h>
#include <io.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//--------------------------------------------------------------------
// define
//--------------------------------------------------------------------

const top_line				=  1;

const line_version		=  3;

const line_file				=  5;
const line_move				=  6;
const line_var				=  7;

const line_space_left	=  9;
const line_space_lib	= 10;
const line_space_info	= 11;

const cursor_column		= 50;

const info_line_1			= 17;
const info_line_2			= 18;
const info_line_3			= 19;
const info_line_A			= 13;
const info_line_B			= 14;

const max_comment			= 70;

const v_down					= 0x80;
const v_right					= 0x40;
const v_comment				= 0x20;
const v_mark					= 0x10;

const table_size			=  8000;
const info_size				= 20000;

const empty						= '.';
const xplayer					= 'x';
const oplayer					= 'o';
const variant					= '#';

const False						= 0;
const True						= 1;

const center					= 8 + 7*16; // Center Intersection
const offset 			 		= 13;
const max_offset 			= 15*offset;

const half 						= 6;

//--------------------------------------------------------------------
// type declaration
//--------------------------------------------------------------------

typedef int boolean;

enum command {cmd_unknown,
							cmd_add_table,
							cmd_button,
							cmd_box_up,
							cmd_box_left,
							cmd_box_right,
							cmd_box_down,
							cmd_box_next,
							cmd_back_one_move,
							cmd_delete_all_marks,
							cmd_find_mark,
							cmd_search_comment,
							cmd_goto_move,
							cmd_change_mode,
							cmd_copy,
							cmd_delete_move,
							cmd_down,
							cmd_edit_comment,
							cmd_mark,
							cmd_save,
							cmd_up,
							cmd_restart,
							cmd_quit,
							cmd_test};

enum mode {display_mode, update_mode};

struct node { unsigned char  pos;			// position
							unsigned char  info;		// information
							unsigned short down;    // next level
							unsigned short right;   // same level
							unsigned short comment;
						};

typedef char comment_line [max_comment+1];

struct comment_string { comment_line line;
												unsigned int length;
											};

typedef char string[100];

//--------------------------------------------------------------------
// data declaration
//--------------------------------------------------------------------

struct node far table[table_size+1];

char far info_buffer[info_size+1];

int the_index;
int old_index;

int box_x;
int box_y;

int box_new_x;
int box_new_y;

int free_list;

char board[240+1];
char g_board[240+1];

int moves[225+1];
int the_move;

int occupied;
int info_start;
int info_left;
int info_garbage;

enum mode the_mode;

FILE *the_file;

string the_pathspec;

string the_file_name;
string the_full_name;
string the_bak_name;

int is_updated;

struct comment_string the_comment;
struct comment_string old_comment;

int the_stack [450];

boolean is_test_on = False;

//--------------------------------------------------------------------
// PC/DOS definitions
//--------------------------------------------------------------------

int gdriver = DETECT, gmode, errorcode;

int current_row;
int current_column;

char buffer[80];

//--------------------------------------------------------------------
// function definitions
//--------------------------------------------------------------------

void put_move (void);
void put_var (boolean Do_Put);
void put_space_left (void);
void put_comment (void);

//--------------------------------------------------------------------

void set_cursor (int row, int column) {
	int height = textheight("*");
	int x = (column-1) * height;
	int y = (row-1) * (2+height) + 4;

	setfillstyle(EMPTY_FILL, getmaxcolor());
	bar(x, y, getmaxx(), y+height);

	current_row    = row;
	current_column = column;
}

//----------------------------------------------------------------------

void set_edit_cursor (int row, int column) {
  current_row    = row;
  current_column = column;
}

//----------------------------------------------------------------------

void print_buffer () {
  int height = textheight("*");
  int x = (current_column-1) * height;
  int y = (current_row-1) * (2 + height) + 4;

	outtextxy (x, y, buffer);
}

//----------------------------------------------------------------------

boolean exists (const char *the_name) {
   struct ffblk ffblk;
	 return (findfirst(the_name, &ffblk, 0) == 0);
}

//----------------------------------------------------------------------

boolean get_dat_bak_file_name(char *file_name,
															char *full_name,
															char *bak_name )
{
	const char_offset = 'A' - 'a';
	int i;

	char *p;

	// make file name uppercase and ending with .DAT

	for (i=0; i < 8; i++) {
		if ((file_name[i] >= 'a') && (file_name[i] <= 'z')) {
			file_name[i] += char_offset;

		} else if ((file_name[i] == '\0') || (file_name[i] == '.')) {
			if (i == 0) {
				return False;
			} else {
				file_name[i] = '.';
				break;
			}

		} else if (((file_name[i] < '0') || (file_name[i] > '9')) &&
							 ((file_name[i] < 'A') || (file_name[i] > 'Z'))) {
			return False;
		}
	}

	file_name[8] = '.'; // Truncate file name, max 8 characters

	p = strchr(file_name, '.') + 1;
	strcpy(p, "DAT");

	// make full file name

	strcpy(full_name, the_pathspec);

	p = strrchr(full_name, '\\');
	p++;

	strcpy(p, file_name); // append file name

	strcpy(bak_name, full_name);

	p = strchr(bak_name, '.');
	p++;

	strcpy(p, "BAK");

	return True;
}

//----------------------------------------------------------------------

boolean delete_bak_file() {

	if (exists(the_bak_name)) {
		set_cursor (top_line, 1);
		sprintf(buffer, "Deleting file: %s", the_bak_name);
		print_buffer();

		return (remove(the_bak_name) == 0);
  } else {
    return (True);
  }
}

//----------------------------------------------------------------------

void rename_dat_file () {

	if (exists(the_full_name) && !exists(the_bak_name)) {

		set_cursor (top_line, 1);
		sprintf(buffer, "Renaming to: %s", the_bak_name);
		print_buffer();

		rename(the_full_name, the_bak_name);
	}
}

//----------------------------------------------------------------------

boolean valid (char ch) {
	return ((ch=='y') || (ch=='Y') || (ch=='n') || (ch=='N'));
}

//----------------------------------------------------------------------

boolean confirmed (const char* the_prompt) {

	char    ch;
	boolean yes;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, the_prompt);
	print_buffer();

  set_cursor (info_line_2, cursor_column);
	sprintf(buffer, "Press Y/N");
  print_buffer();

  do {
    ch = getch();
  } while (!valid(ch));

  set_cursor (info_line_2, cursor_column);

  if ((ch=='y') || (ch=='Y')) {
	 sprintf(buffer, "< Yes >");
	 yes = True;
  } else {
	 sprintf(buffer, "< No >");
    yes = False;
	}
	print_buffer();

  return (yes);
}

//----------------------------------------------------------------------

enum command get_command () {

	const Control_A =  1;
	const Control_C =  3;
	const Control_D =  4;
	const Control_M = 13;
	const Control_Q = 17;
	const Control_R = 18;
	const Control_S = 19;
	const Control_X = 24;

	int  key;
	enum command the_command = cmd_unknown;

	while (the_command == cmd_unknown) {

		//abcdefghijklmnopqrstuvwxyz
		//         1         2
		//12345678901234567890123456

		key = getch();

		//set_cursor (info_line_3, cursor_column);
		//sprintf(buffer, "Key: %d", key);
		//print_buffer();

		switch (key) {
			case Control_A : the_command = cmd_add_table;					break;
			case Control_C : the_command = cmd_copy;							break;
			case Control_D : the_command = cmd_delete_move;				break;
			case Control_M : the_command = cmd_change_mode;				break;
			case Control_Q : the_command = cmd_quit;							break;
			case Control_R : the_command = cmd_restart;						break;
			case Control_S : the_command = cmd_save;							break;
			case Control_X : the_command = cmd_delete_all_marks;	break;

			case 'b' :
			case 'B' : the_command = cmd_back_one_move;						break;

			case 'd' :
			case 'D' : the_command = cmd_down;										break;

			case 'e' :
			case 'E' : the_command = cmd_edit_comment;						break;

			case 'f' :
			case 'F' : the_command = cmd_find_mark;								break;

			case 'g' :
			case 'G' : the_command = cmd_goto_move;								break;

			case 'm' :
			case 'M' : the_command = cmd_mark;										break;

			case 'n' :
			case 'N' : the_command = cmd_box_next;								break;

			case 's' :
			case 'S' : the_command = cmd_search_comment;					break;

			case 'u' :
			case 'U' : the_command = cmd_up;											break;

			case ' ' : the_command = cmd_button;									break;

			case  0 :
				switch (getch()) {
					case  72 : the_command = cmd_box_up;							break;
					case  75 : the_command = cmd_box_left;						break;
					case  77 : the_command = cmd_box_right;						break;
					case  80 : the_command = cmd_box_down;						break;
					case 134 : the_command = cmd_test;								break;
				}
				break;
		}
	}

	return (the_command);
}

//----------------------------------------------------------------------

void calc_info_left () {
	info_left = info_size - info_start + 1 + info_garbage;
}

//----------------------------------------------------------------------

void convert_swe_to_eng (unsigned char& ch) {
	switch (ch) {
		case 134 : ch =  '}'; break; // 
		case 132 : ch =  '{'; break; // 
		case 148 : ch =  '|'; break; // 
		case 143 : ch =  ']'; break; // 
		case 142 : ch =  '['; break; // 
		case 153 : ch = '\\';	break; // 
	}
}

//------------------------------------------------------------------------

void convert_eng_to_swe (unsigned char& ch) {
	switch (ch) {
		case  '}' : ch = 134; break; // 
		case  '{' : ch = 132; break; // 
		case  '|' : ch = 148; break; // 
		case  ']' : ch = 143; break; // 
		case  '[' : ch = 142; break; // 
		case '\\' : ch = 153; break; // 
	}
}

//----------------------------------------------------------------------

void make_uppercase (char* the_string) {
	char *p = the_string;

	while (*p) {
		switch ((unsigned char)*p) {
			case 134 : *p = 143; break; //  => 
			case 132 : *p = 142; break; //  => 
			case 148 : *p = 153; break; //  => 

			default:
				if (*p >= 'a' && *p <= 'z') {
					*p = *p - 'a' + 'A';
				}
		}

		p++;
	}
}

//------------------------------------------------------------------------

void draw_board() {

	int n;
	int from_x, from_y, to_x, to_y;

	from_x = offset;
	from_y = offset;

	to_x = 15*offset;
	to_y = from_y;

	for (n = 1; n <= 15; n++) {
		line(from_x, from_y, to_x, to_y);

		from_y += offset;
		to_y += offset;
	}

	from_x = offset;
	from_y = offset;

	to_x = from_x;
	to_y = 15*offset;

	for (n = 1; n <= 15; n++) {
		line(from_x, from_y, to_x, to_y);

		from_x += offset;
		to_x += offset;
	}

}

//----------------------------------------------------------------------

int calc_position (int pos) {
	if (pos % offset <= 5) {
		return (pos / offset);
	} else if (pos % offset >= 20) {
		return (pos / offset + 1);
	} else {
		return (0);
	}
}

//----------------------------------------------------------------------

void init () {
	int i;

	set_cursor (top_line, 1);
	sprintf(buffer, "RENLIB by Frank Arkbo, Uppsala, Sweden");
	print_buffer();

	for (i = 1; i <= table_size; i++) {
		table[i].pos   = ' ';
		table[i].info   = ' ';
		table[i].down    = i+1; // free list
		table[i].right   = 0;
		table[i].comment = 0;
	}

	table[table_size].down = 0; // end of free list

	free_list = 1;
	occupied  = 0;

	for (i = 1; i <= 240; i++) {
		board[i]   = empty;
		g_board[i] = empty;
	}

	the_index = 0;
	the_move  = 0;

	box_x     = 8;
	box_y     = 8;

	box_new_x = 8;
	box_new_y = 8;

  old_index = 0;

  info_start   = 1;
  info_left    = info_size;
  info_garbage = 0;

  the_mode = display_mode;

	is_updated = False;
}

//------------------------------------------------------------------------

void draw_intersection (int x, int y) {

	int from_x, from_y, to_x, to_y;

  setcolor (BLUE);

  if (x / offset == 1) {
	 from_x = x;
  } else {
	 from_x = x - half;
  }

  from_y = y;

  if (x / offset == 15) {
	 to_x = x;
  } else {
	 to_x = x + half;
  }

	to_y = y;

	line(from_x, from_y, to_x, to_y);

  from_x = x;

  if (y / offset == 1) {
	 from_y = y;
  } else {
	 from_y = y - half;
  }

  to_x = x;

  if (y / offset == 15) {
	 to_y = y;
  } else {
	 to_y = y + half;
  }

	line(from_x, from_y, to_x, to_y);
}

//------------------------------------------------------------------------

void draw_box (int column, int row, int color) {
  int x;
  int y;
  int from_x, from_y, to_x, to_y;

  x = offset * column;
	y = offset * row;

  setcolor (color);

  from_x = x - half;
  from_y = y - half;
  to_x   = x + half;
  to_y   = y - half;
  line(from_x, from_y, to_x, to_y);

  from_x = x - half;
	from_y = y + half;
  to_x   = x + half;
  to_y   = y + half;
  line(from_x, from_y, to_x, to_y);

  from_x = x - half;
  from_y = y - half;
  to_x   = x - half;
  to_y   = y + half;
	line(from_x, from_y, to_x, to_y);

  from_x = x + half;
  from_y = y - half;
  to_x   = x + half;
  to_y   = y + half;
  line(from_x, from_y, to_x, to_y);
}

//------------------------------------------------------------------------

void draw_move (int column, int row, char marker) {
  int x;
  int y;

  x = offset * column;
  y = offset * row;

  setfillstyle(EMPTY_FILL, getmaxcolor());
  bar(x-half, y-half, x+half, y+half);

  setcolor (BLUE);

  if (marker == empty) {
    draw_intersection(x,y);
  } else if (marker == xplayer) {
    // draw black
    setfillstyle(SOLID_FILL, BLUE);
    fillellipse(x, y, half, half);
	} else if (marker == oplayer) {
    // draw white
		setfillstyle(SOLID_FILL, BLUE);
    fillellipse(x, y, half, half);
    setfillstyle(EMPTY_FILL, WHITE);
    fillellipse(x, y, half-1, half-1);
  } else {
    // draw variant
    draw_intersection(x,y);

    if (the_move & 1) {
			setfillstyle(SOLID_FILL, BLUE);
      fillellipse(x, y, 3, 3);
      setfillstyle(EMPTY_FILL, WHITE);
			fillellipse(x, y, 2, 2);
		} else {
			setcolor (BLUE);
			setfillstyle(SOLID_FILL, BLUE);
			fillellipse(x, y, 3, 3);
		}
	}

	setcolor (BLUE);
}

//------------------------------------------------------------------------

void display_board () {
	int i, j, pos;

	// clear board
	for (i = 1; i <= 240; i++) {
		board[i] = empty;
	}

	// set made move markers
	for (i = 1; i <= the_move; i++) {
		pos = table[moves[i]].pos;

		switch (i & 1) {
			case 0 : board[pos] = oplayer; break;
			case 1 : board[pos] = xplayer; break;
		}
	}

	// set new variant markers
	pos = table[the_index].down;

	while (pos) {
		board[table[pos].pos] = variant;
		pos = table[pos].right;
	}

	pos = 1;

	for (i = 1; i <= 15; i++) {
		for (j = 1; j <= 15; j++) {
			if (g_board[pos] != board[pos] || board[pos] == variant) {
				g_board[pos] = board[pos];
				draw_move (j, i, board[pos]);
			}
			pos++;
		}
		pos++;
	}

	draw_box(box_x, box_y, WHITE);
	draw_move (box_x, box_y, board[box_x + 16 * (box_y - 1)]);
	box_x = box_new_x;
	box_y = box_new_y;
	draw_box(box_x, box_y, BLUE);

	if (the_index) {
		pos = table[the_index].pos;
	}
}

//------------------------------------------------------------------------

void set_new_box (int pos) {
	box_new_x = pos % 16;
	box_new_y = pos / 16 + 1;
}

//------------------------------------------------------------------------

void box_next_variant (void) {

	int index = the_index;
	int pos   = 16 * (box_y - 1) + box_x;

	if (table[index].down) {
		index = table[index].down;

		while (True) {
			if (pos < table[index].pos) break;

			if (!table[index].right) {
				index = table[the_index].down;
				break;
			}

			index = table[index].right;
		}

		set_new_box (table[index].pos);
	}
}

//------------------------------------------------------------------------

int msb (char ch) {
	return (ch & 0x80);
}

//------------------------------------------------------------------------

int bit_is_one (int bit_value, char ch) {
	return (ch & bit_value);
}

//------------------------------------------------------------------------

void clear_bit (int bit_value, unsigned char& ch) {
	ch &= ~bit_value;
}

//------------------------------------------------------------------------

void set_bit (int bit_value, unsigned char& ch) {
	ch |= bit_value;
}

//------------------------------------------------------------------------

void make ( int index,
						int pos,
						int information) {

	table[index].pos   = pos;
	table[index].info  = information;
	table[index].down  = 0;
	table[index].right = 0;

	occupied++;
}

//------------------------------------------------------------------------

int get_next_free () {
	int result = 0;

	if (free_list != 0) {
		result = free_list;
		free_list = table[free_list].down;
	}

	return (result);
}

//------------------------------------------------------------------------

int is_variant (int pos) {

	int current;
	int result = 0;

	if (the_index) {
		current = table[the_index].down;

		while (current) {
			if (pos == table[current].pos) {
				result = current;
				break;
			}
			current = table[current].right;
    }
  }

	return (result);
}

//------------------------------------------------------------------------

int is_already_moved (int pos) {

	int result = 0;
	int move;
	int m;

	for (m = 1; m <= the_move; m++) {

		move = table[moves[m]].pos;

		if (pos == move) {
			result = move;
			break;
		}
	}

	return (result);
}

//------------------------------------------------------------------------

void display_move (int pos) {
	int index;

	if (the_move) {
		index = is_variant(pos);

		if (index) {
			the_index       = index;
			the_move++;
			moves[the_move] = the_index;
		}
	}
}

//------------------------------------------------------------------------

void add_move (int index, int new_index) {
	if (table[index].down == 0) {
		table[index].down = new_index;
	} else {
		if (table[new_index].pos < table[table[index].down].pos) {
			table[new_index].right = table[index].down;
			table[index].down      = new_index;
		} else {
			index = table[index].down;

			while (True) {
				if (table[index].right == 0) {
					table[index].right = new_index;
					break;
				} else if (table[new_index].pos < table[table[index].right].pos) {
					table[new_index].right = table[index].right;
					table[index].right     = new_index;
					break;
				}
				index = table[index].right;
			}
		}
	}
}

//------------------------------------------------------------------------

void insert_move (int pos) {
	int index;

	index = is_variant(pos) + is_already_moved(pos);

	if (index) {
		display_move (pos);
		return; // <RETURN>
	} else {
		index = get_next_free();

		if (index) {
			make(index, pos, 0);
			add_move (the_index, index);

			the_index       = index;
			the_move++;
			moves[the_move] = the_index;

		} else {
			set_cursor (info_line_1, cursor_column);
			sprintf(buffer, "NO SPACE: Library is full");
			print_buffer();

			set_cursor (info_line_2, cursor_column);
			sprintf(buffer, "No more moves can be added");
			print_buffer();
		}
	}

	is_updated = True;
}

//------------------------------------------------------------------------

void get_comment (int the_index, boolean convert) {
	int pos;

	the_comment.length = 0;

	if (table[the_index].comment) {
		pos = table[the_index].comment;

		while (True) {
			the_comment.length++;
			the_comment.line[the_comment.length] = info_buffer[pos];

			if (msb(info_buffer[pos])) break;

			pos++;
		}

		if (convert) {
			clear_bit (0x80, (unsigned char)the_comment.line[the_comment.length]);

			for (pos = 1; pos <= the_comment.length; pos++) {
				convert_eng_to_swe ((unsigned char) the_comment.line[pos]);
			}
		}
	}
}

//------------------------------------------------------------------------

void write_comment (int index) {
	int pos;

	unsigned char data1;
	unsigned char data2;

	get_comment (index, False);

	pos = 1;

	while (the_comment.length - pos + 1 >= 2) {
	 data1 = the_comment.line[pos];
	 data2 = the_comment.line[pos+1];
	 pos += 2;

	 fputc (data1, the_file);
	 fputc (data2, the_file);
	}

	if (pos == the_comment.length) {
	 data1 = the_comment.line[pos];
	 data2 = 0;

	 fputc (data1, the_file);
	 fputc (data2, the_file);
	}
}

//------------------------------------------------------------------------

void write_data (int     index,
								 boolean full_tree) {

	// set down bit
	if (table[index].right) {
		set_bit (v_down, table[index].info);
	} else {
		clear_bit (v_down, table[index].info);
	}

	// set right bit
	if (full_tree && !table[index].down) {
		set_bit (v_right, table[index].info);
	} else {
		clear_bit (v_right, table[index].info);
	}

	// set comment bit
	if (table[index].comment) {
		set_bit(v_comment, table[index].info);
	} else {
		clear_bit(v_comment, table[index].info);
	}

	// write data

	fputc (table[index].pos, the_file);
	fputc (table[index].info, the_file);

	// write comment

	if (bit_is_one (v_comment, table[index].info)) {
		write_comment (index);
	}
}

//------------------------------------------------------------------------

int write_node (int move) {

	int m;
	int number	= 0;
	int sp			= 0;

	int index	= table[moves[move]].down;

	boolean done = (index == 0);

	for (m=1; m <= move; m++) {
		write_data (moves[m], False);
		number++;
	}

	while (!done) {
		if (index) {
			write_data (index, True);
			number++;

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
			}

			index = table[index].down;

		} else if (sp) {
			index  = the_stack[--sp];

		} else {
			done = True;
		}
	}

	return (number);
}

//------------------------------------------------------------------------

void write_table (int   move,
									char *file_name,
									char *full_name)
{
	int number;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, file_name);
	print_buffer();

	the_file = fopen (full_name, "wb");

	if (the_file == NULL) {
		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, "ERROR: Failed to open file");
		print_buffer();
		return;
	}

	number = write_node (move);
	fclose (the_file);

	set_cursor (info_line_2, cursor_column);
	sprintf(buffer, "Saved %d moves", number);
	print_buffer();
}

//------------------------------------------------------------------------

void get_string (char *the_string, int min_length, int max_length) {

	int     pos   = 0;
	boolean ready = False;
	char    ch;
	int     c;

	strcpy(the_string, "_");

	do {
		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, the_string);
		print_buffer();

		// wait for key
		c  = getch();
		ch = c;

		set_cursor (info_line_3, cursor_column);

		switch (c) {
			case 8 : // edit_bs
				if (pos > 0) {
					pos--;
					strcpy(&the_string[pos], "_");
				}
				break;

			case  13 : // done
				if (pos >= min_length) {
					ready = True;
				}
				break;

			default :
				if (pos < max_length) {
					if ((c >= 32 && c <= 126) ||
							c == 134 || c == 143 || c == 132 ||
							c == 142 || c == 148 || c == 153) {
						the_string[pos++] = ch;
						strcpy(&the_string[pos], "_");
					}
				} else {
					set_cursor (info_line_3, cursor_column);
					sprintf(buffer, "Max %d characters", max_length);
					print_buffer();
				}
				break;
		}
	} while (!ready);

	the_string[pos] = '\0';
}

//------------------------------------------------------------------------

void copy_command () {

	string file_name;
	string full_name;
	string bak_name;

	if (confirmed("Copy sub-tree to a new file?")) {

		set_cursor (info_line_1, cursor_column);
		sprintf(buffer, "Enter file name:");
		print_buffer();

		while (True) {
			get_string (file_name, 1, 12);

			if (get_dat_bak_file_name (file_name, full_name, bak_name)) {
				break;
			}

			set_cursor (info_line_1, cursor_column);
			sprintf(buffer, "Invalid file name, try again:");
			print_buffer();
		}

		if (exists(full_name)) {
			set_cursor (info_line_1, cursor_column);
			sprintf(buffer, "File %s exists", file_name);
			print_buffer();

			set_cursor (info_line_2, cursor_column);
		} else {
			write_table(the_move, file_name, full_name);
		}
	}
}

//------------------------------------------------------------------------

boolean insert_comment (int index, boolean convert) {
	int  pos;
	int  info_pos;

	// if the comment cannot be appended, return False
	if (info_start + the_comment.length > info_size) return False;

	if (convert) {
		// convert comment to english and set high bit of last character

		for (pos = 1; pos <= the_comment.length; pos++) {
			convert_swe_to_eng ((unsigned char) the_comment.line[pos]);
		}

		set_bit (0x80, (unsigned char)the_comment.line[the_comment.length]);
	}

	if (table[index].comment == 0) {
		// append comment to infobuffer
		table[index].comment = info_start;
		info_start += the_comment.length;
	} else {
		// reduce garbage
		info_garbage = info_garbage - the_comment.length;
	}

	info_pos = table[index].comment;

	for (pos = 1; pos <= the_comment.length; pos++) {
		info_buffer[info_pos] = the_comment.line[pos];
		info_pos++;
	}

	return True;
}

//------------------------------------------------------------------------

void read_comment (void) {

	int data1;
	int data2;

	the_comment.length = 0;

	while (True) {
		data1 = fgetc (the_file);
		data2 = fgetc (the_file);

		the_comment.length++;
		the_comment.line[the_comment.length] = data1;

		if (msb(data1)) break;

		the_comment.length++;
		the_comment.line[the_comment.length] = data2;

		if (msb(data2)) break;
	}
}

//------------------------------------------------------------------------

void find_index (int index_to_find) {

	const current_move = the_move;

	int index	 = table[the_index].down;
	int sp		 = 0;

	boolean done = False;

	while (!done) {
		if (index) {
			the_index = index;
			moves[++the_move] = the_index;

			if (index == index_to_find) return;

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
				the_stack[sp++] = the_move;
			}

			index = table[index].down;

		} else if (sp) {
			the_move = the_stack[--sp];
			index    = the_stack[--sp];

			the_index = moves[the_move--];

		} else {
			done = True;
		}
	}

	the_move  = current_move;
	the_index = moves[the_move];
}

//------------------------------------------------------------------------

int get_variant (int index, int pos) {

	if (table[index].down) {
		index = table[index].down;

		if (table[index].pos == pos) return index;

		while (table[index].right) {
			index = table[index].right;
			if (table[index].pos == pos) return index;
		}
	}

	return 0;
}

//------------------------------------------------------------------------

void add_table (char *file_name,
								char *full_name) {

	int number = 0;
	int sp     = 0;

	int first_index	= 0;
	int next_index;
	int data;

	int no_of_comments	= 0;
	int no_of_marks			= 0;

	int no_of_new_moves					= 0;
	int no_of_old_moves					= 1;
	int no_of_skipped_moves			= 0;
	int no_of_skipped_comments	= 0;

	struct ftime ft;

	struct node next;

	boolean is_new = False;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, file_name);
	print_buffer();

	the_file = fopen (full_name, "rb");

	if (the_file == NULL) {
		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, "ERROR: Failed to open file");
		print_buffer();
		return;
	}

	getftime(fileno(the_file), &ft);

	while ((data = fgetc (the_file)) != EOF) {
		number++;
		next.pos  = data;
		next.info = fgetc(the_file);

		if (number == 1) {
			if (the_index == 0) {
				the_index = get_next_free();

				make(the_index, center, next.info);

				the_move        = 1;
				moves[the_move] = the_index;

				first_index     = the_index;
				no_of_new_moves = 1;

				is_new = True;
			} else {
				the_move  = 1;
				the_index = moves[the_move];
			}
		} else {
			next_index = get_variant(the_index, next.pos);

			if (next_index) {
				the_index = next_index;

				no_of_old_moves++;
				is_new = False;

			} else {
				next_index = get_next_free();

				if (next_index) {
					is_new = True;

					make(next_index, next.pos, next.info);
					add_move (the_index, next_index);
					the_index = next_index;

					if (first_index == 0) first_index = the_index;
					no_of_new_moves++;
				} else {
					no_of_skipped_moves++;
					is_new = False;
				}
			}

			moves[++the_move] = the_index;
		}

		if (bit_is_one (v_comment, next.info)) {
			read_comment ();

			if (is_new) {
				if (insert_comment (the_index, False)) {
					no_of_comments++;
				} else {
					no_of_skipped_comments++;
				}
			}
		}

		if (bit_is_one(v_mark, table[number].info) && is_new) {
			no_of_marks++;
		}

		if (bit_is_one(v_down, next.info)) {
			the_stack[sp++] = the_move;
		}

		if (bit_is_one(v_right, next.info)) {
			if (sp) {
				the_move    = the_stack[--sp];
				the_index   = moves[--the_move];
			}
		}
	}

	fclose (the_file);

	the_index       = 1;
	the_move        = 1;
	moves[the_move] = the_index;
	old_index       = 0;

	set_new_box (center);

	set_cursor (info_line_2, cursor_column);
	if (no_of_new_moves == number) {
		sprintf(buffer, "Read %d moves", number);
	} else {
		sprintf(buffer, "Added %d of %d moves",
										no_of_new_moves, number);

		if (no_of_new_moves) {
			is_updated = True;
		}

		if (first_index) {
			find_index (first_index);
			set_new_box (table[the_index].pos);
		}
	}
	print_buffer();

	set_cursor (info_line_3, cursor_column);
	sprintf(buffer,"%d-%02d-%02d %02d:%02d:%02d",
									ft.ft_year+1980, ft.ft_month, ft.ft_day,
									ft.ft_hour, ft.ft_min, ft.ft_tsec * 2);
	print_buffer();

	if (no_of_skipped_moves + no_of_skipped_comments) {
		if (no_of_skipped_moves) {
			set_cursor (info_line_A, cursor_column);
			sprintf(buffer, "NO SPACE: Skipped %d moves", no_of_skipped_moves);
			print_buffer();
		}

		if (no_of_skipped_comments) {
			set_cursor (info_line_B, cursor_column);
			sprintf(buffer, "NO SPACE: Skipped %d comments", no_of_skipped_comments);
			print_buffer();
		}

	} else {
		set_cursor (info_line_A, cursor_column);
		sprintf(buffer, "No of comments : %d", no_of_comments);
		print_buffer();

		set_cursor (info_line_B, cursor_column);
		sprintf(buffer, "No of marks    : %d", no_of_marks);
		print_buffer();
	}
}

//------------------------------------------------------------------------

void add_table_command () {

	string file_name;
	string full_name;
	string bak_name;

	if (!confirmed("Add table from file?")) return;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, "Enter file name:");
	print_buffer();

	while (True) {
		get_string (file_name, 1, 12);

		if (get_dat_bak_file_name (file_name, full_name, bak_name)) {
			if (exists(full_name)) break;

			set_cursor (info_line_1, cursor_column);
			sprintf(buffer, "File does not exist");
			print_buffer();
		} else {
			set_cursor (info_line_1, cursor_column);
			sprintf(buffer, "Invalid file name, try again:");
			print_buffer();
		}
	}

	add_table (file_name, full_name);
}

//------------------------------------------------------------------------

void read_table () {

	add_table (the_file_name, the_full_name);

}

//------------------------------------------------------------------------

int count_total (void) {

	int number = 0;
	int index	 = table[the_index].down;
	int sp		 = 0;

	boolean done = False;

	while (!done) {
		if (index) {
			number++;

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
			}

			index = table[index].down;

		} else if (sp) {
			index  = the_stack[--sp];

		} else {
			done = True;
		}
	}

	return (number);
}

//------------------------------------------------------------------------

void number_of_variants (int& local, int& total) {
	int index;

	total = count_total ();

	local = 0;

	if (table[the_index].down) {
		local++;

		index = table[the_index].down;

		while (table[index].right) {
			local++;
			index = table[index].right;
		}
	}
}

//------------------------------------------------------------------------

void back_command () {
	int pos;

	if (the_move > 1) {
		pos = table[the_index].pos;
		board[pos] = empty;

		set_new_box (pos);

		the_move--;
		the_index = moves[the_move];
	}
}

//------------------------------------------------------------------------

void delete_comment (int index, boolean clear) {
	int pos;

	if (table[index].comment) {
		// make previous space free

		pos = table[index].comment;

		while (!msb(info_buffer[pos])) {
			info_buffer[pos] = 0;
			pos++;
		}

		info_buffer[pos] = 0;

		info_garbage = info_garbage + pos - table[index].comment + 1;

		if (clear) {
			table[index].comment = 0;

			while (info_start >= 2) {
				if (info_buffer[info_start - 1]) break;

				info_start   = info_start - 1;
				info_garbage = info_garbage - 1;
			}
		}
	}
}

//------------------------------------------------------------------------

void remove_move (int index) {

  if (index) {
	 remove_move (table[index].down);
	 remove_move (table[index].right);

	 table[index].down  = free_list;
	 table[index].right = 0;
	 free_list          = index;
	 occupied           = occupied - 1;

	 delete_comment (index, True);
	}
}

//------------------------------------------------------------------------

void search_comment (int comment, int index, int found) {
	if (table[index].comment == comment) {
	 found = index;
	} else {
	 if (!found && table[index].down) {
		search_comment (comment, table[index].down, found);
	 }

	 if (!found && table[index].right) {
		search_comment (comment, table[index].right, found);
	 }
	}
}

int target_comment (int comment) {
	int found = 0;

	search_comment (comment, 1, found);
	return (found);
}

//------------------------------------------------------------------------

int garbage_collect (int space) {

	int i;
	int pos = 0;
	int pos1;
	int pos2;
	int pos3;
	int length;
	int index;
	int result = 0;

	boolean done = False;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, "Garbage Collect - wait...");
	print_buffer();

	while (!done) {
		do {
			pos++;
			if (pos == info_start) {
				done = True;
				break;
			}
		} while (info_buffer[pos]);


		if (!done) {
			pos1 = pos;

			do {
				pos++;
				if (pos == info_start) {
					done = True;
					break;
				}
			} while (!info_buffer[pos]);
		}


		if (!done) {
			pos2   = pos;
			length = pos2 - pos1;

			if (length >= space) {
				result = pos1;
				done = True; // required space is available
				break;
			}
		}


		if (!done) {
			while (!msb(info_buffer[pos])) {
				pos++;
			}

			pos3 = pos;

			index = target_comment (pos2);

			table[index].comment = pos1;

			for (i = pos2; i <= pos3; i++) {
				info_buffer[i - length] = info_buffer[i];
			}

			for (i = pos3 - length + 1; i <= pos3; i++) {
				info_buffer[i] = 0;
			}

			pos = pos3 - length;
		}
	}

	if (result == 0) {
		while (!info_buffer[info_start - 1]) {
			info_start = info_start - 1;
		}

		info_garbage = 0;
	}

	return (result);
}

//------------------------------------------------------------------------

void delete_command () {
	int     local, total;
	int     index;
	char    question[30];
	boolean OK_to_delete = True;

	if (the_move == 1) {
		set_cursor (info_line_1, cursor_column);
		sprintf(buffer, "Can't delete first move!");
		print_buffer();
	} else {
		number_of_variants (local, total);

		if (total > 0) {
			sprintf(question, "Delete %d moves ?", total + 1);

			if (confirmed(question) && confirmed("Are you really sure ?")) {
				remove_move(table[the_index].down);
			} else {
				OK_to_delete = False;
			}
		}

		if (OK_to_delete) {
			index = moves[the_move-1];

			// Unlink the current move

			if (table[index].down == the_index) {
				table[index].down = table[the_index].right;
			} else {
				index = table[index].down;

				while (index) {
					if (table[index].right == the_index) {
						table[index].right = table[the_index].right;
						break;
					}

					index = table[index].right;
				}
			}

			board[table[the_index].pos] = empty;

			table[the_index].down = 0;
			table[the_index].right = 0;

			remove_move(the_index);

			the_move--;
			the_index = moves[the_move];

			is_updated = True;
		}
	}
}

//------------------------------------------------------------------------

void save_command () {

	delete_bak_file();
	// rename_dat_file(); // optional ???

	write_table (1, the_file_name, the_full_name);

	is_updated = False;
}

//------------------------------------------------------------------------

void delete_char (int pos) {
	int i;

	for (i = pos + 1; i <= the_comment.length; i++) {
	 the_comment.line[i-1] = the_comment.line[i];
	}

	the_comment.length = the_comment.length - 1;
}

//------------------

void insert_char (char ch, int pos) {
	int i;

	for (i = the_comment.length; i >= pos; i--) {
	 the_comment.line[i+1] = the_comment.line[i];
	}

	the_comment.line[pos] = ch;
	the_comment.length++;
}

//------------------

void word_left (int& pos) {
	if (pos > 1 && the_comment.line[pos - 1] == ' ') {
		pos--;
	}

	while (pos > 1 && the_comment.line[pos] == ' ') {
		pos--;
	}

	while (pos > 1 && the_comment.line[pos - 1] != ' ') {
		pos--;
	}
}

//------------------

void word_right (int& pos) {
	if (pos <= the_comment.length && the_comment.line[pos + 1] == ' ') {
		pos++;
	}

	while (pos <= the_comment.length && the_comment.line[pos] != ' ') {
		pos++;
	}

	while (pos <= the_comment.length && the_comment.line[pos] == ' ') {
		pos++;
	}
}

//------------------

void comment_command () {

	int     pos;
	boolean ready;
	char    ch;
	int     c;
	int     max_length;
	int     garbage;

	// check max length of comment

	if (table[the_index].comment == 0 && info_left == 0) {
		set_cursor (info_line_1, cursor_column);
		sprintf(buffer, "No remaining space!");
		print_buffer();
		return;
	}

	get_comment (the_index, True);

	// save comment in old_comment
	strncpy(&old_comment.line[1], &the_comment.line[1], the_comment.length);
	old_comment.length = the_comment.length;

	garbage = 0;

	if (table[the_index].comment) {
		// count local garbage
		pos = table[the_index].comment;

		while (!msb(info_buffer[pos])) {
			pos++;
		}

		pos++;

		while (pos != info_start) {
			if (info_buffer[pos]) break;

			garbage++;
			pos++;
		}
	}

	if (info_left + old_comment.length < max_comment) {
		max_length = info_left + old_comment.length;
	} else {
		max_length = max_comment;
	}

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, "<Edit>   Max   : %d", max_length);
	print_buffer();

	ready = False;
	pos   = 1;

	do {
		the_comment.line[0] = '\"';
		set_cursor (top_line, 1);
		strcpy(buffer, "E: ");
		strncat(buffer, the_comment.line, the_comment.length + 1);
		strcat(buffer, "\"");
		print_buffer();

		// display <cursor>
		set_edit_cursor (top_line, pos + 4);
		sprintf(buffer, "%c", 177);
		print_buffer();

		// show info
		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, "Pos: %2d, Length: %2d", pos, the_comment.length);
		print_buffer();

		// wait for key
		c  = getch();
		ch = c;

		if (c >= 32 && c <= 126) { // characters
			// edit_ch
			if (the_comment.length < max_length) {
				convert_eng_to_swe ((unsigned char) ch);
				insert_char (ch, pos);
				pos++;
			}
		} else { // special keys on keyboard
			switch (c) {
				case   0 :
					// get following ch
					c  = getch();
					ch = c;

					switch (c) {
						case  75 : // edit_left
							if (pos > 1) {
								pos--;
							}
							break;

						case  77 : // edit_right
							if (pos <= the_comment.length) {
								pos++;
							}
							break;

						case  82 : // edit_insert
							strncpy(&the_comment.line[1], &old_comment.line[1],
											old_comment.length);
							the_comment.length = old_comment.length;
							pos = 1;
							break;

						case 83 : // edit_delete
							if (pos <= the_comment.length) {
								delete_char (pos);
							}
							break;

						case  71 : // edit_home
							pos = 1;
							break;

						case  79 : // edit_end
							pos = the_comment.length + 1;
							break;

						case 115 : // edit_control_left
							word_left(pos);
							break;

						case 116 : // edit_control_right
							word_right(pos);
							break;

						default  :
							break;
					}
					break;

				case   8 : // edit_backspace
					if (pos > 1) {
						pos = pos - 1;
						delete_char (pos);
					}
					break;

				case  13 : // done
					ready = True;
					break;

				case  27 : // edit_esc
					the_comment.length = 0;
					pos                = 1;
					break;

				case 132 : // '{'
				case 134 : // '}'
				case 142 : // '['
				case 143 : // ']'
				case 148 : // '|'
				case 153 : // '\\'
					// edit_ch
					if (the_comment.length < max_length) {
						insert_char (ch, pos);
						pos++;
					}
					break;

				default :
					break;
			}
		}
	} while (!ready);

	// strip leading blanks
	while (the_comment.length && (the_comment.line[1] == ' ')) {
	 delete_char (1);
	}

	// strip trailing blanks
	while (the_comment.length && (the_comment.line[the_comment.length] == ' ')) {
	 delete_char (the_comment.length);
	}

	// check if updated etc.
	set_cursor (info_line_2, cursor_column);
	set_cursor (info_line_1, cursor_column);

	if (the_comment.length == 0) {
		if (old_comment.length) {
			sprintf(buffer, "Deleted comment");
			print_buffer();
			is_updated = True;
		} else {
			sprintf(buffer, "No comment");
			print_buffer();
		}
	} else if (old_comment.length == 0) {
		sprintf(buffer, "Added comment");
		print_buffer();
		is_updated = True;
	} else if (strncmp(&the_comment.line[1], &old_comment.line[1],
											the_comment.length)) {
		sprintf(buffer, "Updated comment");
		print_buffer();
		is_updated = True;
	} else {
		sprintf(buffer, "Unchanged comment");
		print_buffer();
	}

	// take care of the comment
	if (the_comment.length == 0) {
		// no comment
		delete_comment (the_index, True);

	} else {
		// check local garbage

		if (old_comment.length > 0 &&
			the_comment.length <= old_comment.length + garbage) {
			// use same space again
			delete_comment (the_index, False);

		} else {
			// append comment
			delete_comment (the_index, True);

			calc_info_left();

			if (the_comment.length > info_left - info_garbage) {
				table[the_index].comment = garbage_collect (the_comment.length);
			}
		}

		if (!insert_comment (the_index, True)) {
			set_cursor (info_line_3, cursor_column);
			sprintf(buffer, "ERROR: Failed to add comment");
			print_buffer();
		}
	}
}

//------------------------------------------------------------------------

void set_position (int column, int row) {
	int pos;

	pos = 16 * (row - 1) + column;

	switch (the_mode) {
		case update_mode  : insert_move (pos);  break;
		case display_mode : display_move (pos); break;
	}

	box_new_x = column;
	box_new_y = row;
}

//----------------------------------------------------------------------

void notify_bad_mode () {
	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, "Not in Update Mode");
	print_buffer();
}

//----------------------------------------------------------------------

boolean get_single_variant (int& column, int& row) {
	int index;

	// first variant
	index = table[the_index].down;

	if (index && !table[index].right) {
		column = table[index].pos % 16;
		row    = table[index].pos / 16 + 1;
		return True;
	} else {
		column = 0;
		row    = 0;
		return False;
	}
}

//----------------------------------------------------------------------

boolean is_comment (int index) {
	return table[index].comment;
}

//----------------------------------------------------------------------

boolean is_mark (int index) {
	return bit_is_one (v_mark, table[index].info);
}

//----------------------------------------------------------------------

void set_mark (int index) {
	set_bit(v_mark, table[index].info);
}

//----------------------------------------------------------------------

void clear_mark (int index) {
	clear_bit(v_mark, table[index].info);
}

//----------------------------------------------------------------------

boolean is_comment_found (const char* search_string, int index) {

	string comment;

	get_comment (index, True);

	if (strlen(search_string) > the_comment.length) {
		return False;
	}

	strncpy(comment, &the_comment.line[1], the_comment.length);
	comment[the_comment.length] = '\0';
	make_uppercase(comment);
	return strstr(comment, search_string) != NULL;
}

//----------------------------------------------------------------------

void search_comment_command (void) {

	string search_string;

	const current_move = the_move;

	int index	 = table[the_index].down;
	int sp		 = 0;

	boolean search_all = False;

	boolean done = False;

	set_cursor (info_line_1, cursor_column);
	sprintf(buffer, "Enter string to search:");
	print_buffer();

	get_string (search_string, 0, 20);

	if (strlen(search_string) == 0) {
		if (confirmed("Search ALL comments?")) {
			search_all = True;
		} else {
			return;
		}
	} else {
		set_cursor (info_line_3, cursor_column);
		sprintf(buffer, "Search: \"%s\"", search_string);
		print_buffer();

		make_uppercase(search_string);
	}

	while (!done) {
		if (index) {
			the_index = index;
			moves[++the_move] = the_index;

			if (is_comment(index)) {
				if (search_all || is_comment_found (search_string, index)) {
					display_board();
					put_move();
					put_var (False);
					put_comment();

					if (!confirmed("Search next comment?")) {
						set_cursor (info_line_3, cursor_column);
						return;
					}
				}
			}

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
				the_stack[sp++] = the_move;
			}

			index = table[index].down;

		} else if (sp) {
			the_move = the_stack[--sp];
			index    = the_stack[--sp];

			the_index = moves[the_move--];

		} else {
			done = True;
		}
	}

	the_move  = current_move;
	the_index = moves[the_move];

	set_cursor (info_line_1, cursor_column);
	set_cursor (info_line_2, cursor_column);
	sprintf(buffer, "No more comments");
	set_cursor (info_line_3, cursor_column);
	print_buffer();
}

//------------------------------------------------------------------------

int count_delete_marks (boolean delete_mark) {

	int index	 = table[the_index].down;
	int sp		 = 0;

	boolean done = False;

	int n_marks = 0;

	if (is_mark(the_index)) {
		n_marks++;

		if (delete_mark) {
			clear_mark (the_index);
		}
	}

	while (!done) {
		if (index) {
			if (is_mark(index)) {
				n_marks++;

				if (delete_mark) {
					clear_mark (index);
				}
			}

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
			}

			index = table[index].down;

		} else if (sp) {
			index = the_stack[--sp];

		} else {
			done = True;
		}
	}

	return n_marks;
}

//------------------------------------------------------------------------

void delete_all_marks_command (void) {

	char	question[30];
	int		n_marks = count_delete_marks (False);


	if (n_marks > 0) {
		sprintf(question, "Delete %d marks ?", n_marks);

		if (confirmed(question)) {
			count_delete_marks (True);
		}
	} else {
		set_cursor (info_line_1, cursor_column);
		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, "No marks");
		print_buffer();
	}
}

//------------------------------------------------------------------------

void find_mark_command (void) {

	const current_move = the_move;

	int index	 = table[the_index].down;
	int sp		 = 0;

	boolean done = False;

	while (!done) {
		if (index) {
			the_index = index;
			moves[++the_move] = the_index;

			if (is_mark(index)) {
				display_board();
				put_move();
				put_var (False);
				put_comment();

				if (!confirmed("Find next mark?")) {
					return;
				}
			}

			if (table[index].right) {
				the_stack[sp++] = table[index].right;
				the_stack[sp++] = the_move;
			}

			index = table[index].down;

		} else if (sp) {
			the_move = the_stack[--sp];
			index    = the_stack[--sp];

			the_index = moves[the_move--];

		} else {
			done = True;
		}
	}

	the_move  = current_move;
	the_index = moves[the_move];

	set_cursor (info_line_1, cursor_column);
	set_cursor (info_line_2, cursor_column);
	sprintf(buffer, "No more marks");
	print_buffer();
}

//------------------------------------------------------------------------

void mark_command (void) {

	if (is_mark(the_index)) {
		clear_mark (the_index);

		set_cursor (info_line_2, cursor_column);
		sprintf(buffer, "MARK cleared");
		print_buffer();
	} else {
		set_mark (the_index);
	}

	is_updated = True;
}

//----------------------------------------------------------------------

void down_command (void) {
	int row;
	int column;

	while (True) {
		if (get_single_variant (column, row)) {
			set_position (column, row);
		} else {
			break;
		}
	}
}

//----------------------------------------------------------------------

void initialize_graphics () {
	// initialize graphics and local variables
	initgraph(&gdriver, &gmode, "");

	 // read result of initialization
	 errorcode = graphresult();

	if (errorcode != grOk) {
		printf("Graphics error:\n%s\n", grapherrormsg(errorcode));
		printf("Press any key to halt:");
		getch();
		exit(1);
	}

	//if (getmaxx() < max_offset || getmaxy() < max_offset) {
	//  closegraph();
	//  printf("Graphic screen is smaller than %d x %d", max_offset, max_offset);
	//  printf("\npress <RETURN> to exit");
	//  getch();
	//  return;
	//}

	setbkcolor(LIGHTGRAY);
	setcolor(BLUE);
}

//------------------------------------------------------------------------

char *current_directory (char *path) {
	strcpy(path, "X:\\");			// fill string with form of response: X:\
	path[0] = 'A' + getdisk();	// replace X with current drive letter
	getcurdir(0, path+3);			// fill rest of string with current directory
	return(path);
}

//------------------------------------------------------------------------

void directory () {
	struct find_t ffblk;
	char name[13];         // found file name
	int done;
	int n=0;

	current_directory(the_pathspec);

	if (the_pathspec[strlen(the_pathspec)-1] != '\\') {
		strcat(the_pathspec, "\\");
	}

	strcat(the_pathspec, "*.DAT");

	printf("Directory listing of %s\n", the_pathspec);
	done = _dos_findfirst(the_pathspec, _A_NORMAL, &ffblk);

	while (!done) {
		if (n==0) printf("\n");
		strcpy(name, ffblk.name);
		name[strlen(name)-4]='\0';
		printf("%-9s", name);
		done = _dos_findnext(&ffblk);
		n = (n + 1) % 8;
	}
	printf("\n\n");
}

//------------------------------------------------------------------------

void author ( ) {

	int number[36] = {0, 51, 49, 36, 42, 44, -13, -40, -3, 40, 22, 34,
			 30, -46, -14, 34, 26, 16, 28, -40, -53, -1, 25,
			 24, 26, 7, 17, 5, -49, -62, -12, 23, 4, 2, 2, 10};

	int c = 'A';
	int i = 0;

	clrscr();

	printf("-----------------------------------------------------------------");
	printf("\nWelcome to RENLIB\n");

	while (i <= 35) {
		printf("%c", number[i]+c);
		c++;
		i++;
	}

	printf("\n");
	printf("-----------------------------------------------------------------");
	printf("\n\n");
}

//------------------------------------------------------------------------

void put_move (void) {
	set_cursor (line_move, cursor_column);
	sprintf(buffer, "Move : %2d", the_move);
	print_buffer();

	if (is_mark(the_index)) {
		set_cursor (line_move, cursor_column + 12);
		sprintf(buffer, "*** MARK ***");
		print_buffer();
	}
}

//------------------------------------------------------------------------

void put_var (boolean Do_Put) {
	int local, total;

	if (!Do_Put) {
		set_cursor (line_var, cursor_column);
	} else if (old_index != the_index) {
		set_cursor (line_var, cursor_column);
		number_of_variants (local, total);
		sprintf(buffer, "Var  : %2d of %d", local, total);
		print_buffer();

		old_index = the_index;
	}
}

//------------------------------------------------------------------------

void put_space_left (void) {

	set_cursor (line_space_left, cursor_column);
	if (the_mode == update_mode) {
		sprintf(buffer, "SPACE LEFT");
		print_buffer();
	}

	set_cursor (line_space_lib, cursor_column);
	if (the_mode == update_mode) {
		sprintf(buffer, "Lib  : %5d", table_size - occupied);
		print_buffer();

		set_cursor (line_space_lib, cursor_column + 15);
		if (is_updated) {
			sprintf(buffer, "Updated");
			print_buffer();
		}
	}

	set_cursor (line_space_info, cursor_column);
	if (the_mode == update_mode) {
		sprintf(buffer, "Info : %5d", info_left);
		print_buffer();
	}
}

//------------------------------------------------------------------------

void put_comment (void) {

	get_comment (the_index, True);

	set_cursor (top_line, 1);
	if (the_comment.length > 0) {
		the_comment.line[0] = ' ';
		strcpy(buffer, "C:");
		strncat(buffer, the_comment.line, the_comment.length + 1);
		print_buffer();
	}
}

//------------------------------------------------------------------------

void put_test (void) {

	if (is_test_on) {
		set_cursor (info_line_A, cursor_column);
		sprintf(buffer, "the_index : %3d", the_index);
		print_buffer();

		set_cursor (info_line_B, cursor_column);
		sprintf(buffer, "pos       : %3d", table[the_index].pos);
		print_buffer();
	}
}

//------------------------------------------------------------------------

void execute (boolean& again) {
	enum command the_command;

	struct time time;

	int local, total;
	int row, column;

	boolean ready      = False;

	again = False;

	down_command();

	while (!ready) {
		display_board();

		calc_info_left();

		set_cursor (line_file, cursor_column);
		sprintf(buffer, "File : %s", the_file_name);
		print_buffer();

		put_move();
		put_var (True);
		put_space_left();
		put_comment();
		put_test();

		// wait for command

		the_command = get_command();

		set_cursor (info_line_1, cursor_column);

		set_cursor (info_line_2, cursor_column);
		gettime(&time);
		sprintf(buffer, "Time: %2d:%02d:%02d",
										time.ti_hour, time.ti_min, time.ti_sec);
		print_buffer();

		set_cursor (info_line_3, cursor_column);

		set_cursor (info_line_A, cursor_column);
		set_cursor (info_line_B, cursor_column);

		switch (the_command) {
			case cmd_add_table :
				if (the_mode == update_mode) {
					add_table_command();
				} else {
					notify_bad_mode();
				}
				break;

			case cmd_button :
				if (the_mode == display_mode &&
						get_single_variant (column, row)) {

					set_position (column, row);
				} else {
					set_position (box_x, box_y);
				}
				break;

			case cmd_box_up :
				box_new_y--;

				if (box_new_y < 1) {
					box_new_y = 15;
				}
				break;

			case cmd_box_left :
				box_new_x--;

				if (box_new_x < 1) {
					box_new_x = 15;
				}
				break;

			case cmd_box_right :
				box_new_x++;

				if (box_new_x > 15) {
					box_new_x = 1;
				}
				break;

			case cmd_box_down :
				box_new_y++;

				if (box_new_y > 15) {
					box_new_y = 1;
				}
				break;

			case cmd_box_next :
				box_next_variant();
				break;

			case cmd_back_one_move :
				back_command();
				break;

			case cmd_delete_all_marks :
				if (the_mode == update_mode) {
					delete_all_marks_command();
				} else {
					notify_bad_mode();
				}
				break;

			case cmd_find_mark :
				find_mark_command();
				break;

			case cmd_search_comment :
				search_comment_command();
				break;

			case cmd_goto_move :
				while (the_move > 1) {
					back_command();
					if (is_mark(the_index)) break;
				}
				break;

			case cmd_change_mode :
				switch (the_mode) {
					case update_mode  :
						the_mode = display_mode;
						set_cursor (info_line_1, cursor_column);
						sprintf(buffer, "Display Mode");
						print_buffer();
						break;

					case display_mode :
						the_mode = update_mode;
						set_cursor (info_line_1, cursor_column);
						sprintf(buffer, "Update Mode");
						print_buffer();
						break;
				}
				break;

			case cmd_copy :
				copy_command();
				break;

			case cmd_delete_move :
				if (the_mode == update_mode) {
					delete_command();
				} else {
					notify_bad_mode();
				}
				break;

			case cmd_down :
				down_command();
				break;

			case cmd_edit_comment :
				if (the_mode == update_mode) {
					comment_command();
				} else {
					notify_bad_mode();
				}
				break;

			case cmd_mark :
				if (the_mode == update_mode) {
					mark_command();
				} else {
					notify_bad_mode();
				}
				break;

			case cmd_save :
				save_command();
				break;

			case cmd_up :
				while (the_move > 1) {
					back_command();
					number_of_variants (local, total);
					if (local > 1) break;
				}
				break;

			case cmd_restart :
			case cmd_quit    :
				{
					if (is_updated && confirmed("Save data?")) {
						save_command();
					}

					again = (the_command == cmd_restart);
					ready  = True;
					break;
				}
			case cmd_test :
				is_test_on = !is_test_on;
				break;
		}
	}
}

//------------------------------------------------------------------------

void start () {

	boolean again;
	boolean ok    = True;

	author();

	do {
		directory();

		while (True) {
			printf("Enter file name: ");
			gets(the_file_name);

			if (get_dat_bak_file_name (the_file_name, the_full_name, the_bak_name)) {
				break;
			}

			printf("*** Invalid file name, try again\n");
		}

		initialize_graphics();
		cleardevice();

		init();

		if (ok) {
			if (exists(the_full_name)) {
				read_table();
			} else {
				insert_move (center);
				save_command();

				set_cursor (info_line_2, cursor_column);
				sprintf(buffer, "New file");
				print_buffer();
			}

			draw_board();

			set_cursor (line_version, cursor_column);
			sprintf(buffer, version);
			print_buffer();

			execute (again);
		} else {
			again = False;
		}

		cleardevice();
		closegraph();

	} while (again);
}

//------------------------------------------------------------------------

void main () {
	start();
}
